package sun.corba;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Field;
import java.security.AccessController;
import java.io.ObjectInputStream;
import java.security.PrivilegedAction;
import sun.reflect.ReflectionFactory;
import sun.misc.Unsafe;
import java.lang.reflect.Method;
import java.security.Permission;

public final class Bridge
{
    private static final Class[] NO_ARGS;
    private static final Permission getBridgePermission;
    private static Bridge bridge;
    private final Method latestUserDefinedLoaderMethod;
    private final Unsafe unsafe;
    private final ReflectionFactory reflectionFactory;
    public static final long INVALID_FIELD_OFFSET = -1L;
    
    private Method getLatestUserDefinedLoaderMethod() {
        return (Method)AccessController.doPrivileged(new PrivilegedAction() {
            @Override
            public Object run() {
                Method declaredMethod;
                try {
                    declaredMethod = ObjectInputStream.class.getDeclaredMethod("latestUserDefinedLoader", Bridge.NO_ARGS);
                    declaredMethod.setAccessible(true);
                }
                catch (NoSuchMethodException ex) {
                    Error error = new Error("java.io.ObjectInputStream latestUserDefinedLoader " + ex);
                    error.initCause(ex);
                    throw error;
                }
                return declaredMethod;
            }
        });
    }
    
    private Unsafe getUnsafe() {
        Field field = (Field)AccessController.doPrivileged(new PrivilegedAction() {
            @Override
            public Object run() {
                try {
                    Field declaredField = Unsafe.class.getDeclaredField("theUnsafe");
                    declaredField.setAccessible(true);
                    return declaredField;
                }
                catch (NoSuchFieldException cause) {
                    Error error = new Error("Could not access Unsafe");
                    error.initCause(cause);
                    throw error;
                }
            }
        });
        Unsafe unsafe;
        try {
            unsafe = (Unsafe)(Unsafe)field.get(null);
        }
        catch (Throwable cause) {
            Error error = new Error("Could not access Unsafe");
            error.initCause(cause);
            throw error;
        }
        return unsafe;
    }
    
    private Bridge() {
        this.latestUserDefinedLoaderMethod = this.getLatestUserDefinedLoaderMethod();
        this.unsafe = this.getUnsafe();
        this.reflectionFactory = (ReflectionFactory)AccessController.doPrivileged(new ReflectionFactory.GetReflectionFactoryAction());
    }
    
    public static final synchronized Bridge get() {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkPermission(Bridge.getBridgePermission);
        }
        if (Bridge.bridge == null) {
            Bridge.bridge = new Bridge();
        }
        return Bridge.bridge;
    }
    
    public final ClassLoader getLatestUserDefinedLoader() {
        try {
            return (ClassLoader)this.latestUserDefinedLoaderMethod.invoke(null, (Object[])Bridge.NO_ARGS);
        }
        catch (InvocationTargetException ex) {
            Error error = new Error("sun.corba.Bridge.latestUserDefinedLoader: " + ex);
            error.initCause(ex);
            throw error;
        }
        catch (IllegalAccessException ex2) {
            Error error2 = new Error("sun.corba.Bridge.latestUserDefinedLoader: " + ex2);
            error2.initCause(ex2);
            throw error2;
        }
    }
    
    public final int getInt(Object o, long offset) {
        return this.unsafe.getInt(o, offset);
    }
    
    public final void putInt(Object o, long offset, int x) {
        this.unsafe.putInt(o, offset, x);
    }
    
    public final Object getObject(Object o, long offset) {
        return this.unsafe.getObject(o, offset);
    }
    
    public final void putObject(Object o, long offset, Object x) {
        this.unsafe.putObject(o, offset, x);
    }
    
    public final boolean getBoolean(Object o, long offset) {
        return this.unsafe.getBoolean(o, offset);
    }
    
    public final void putBoolean(Object o, long offset, boolean x) {
        this.unsafe.putBoolean(o, offset, x);
    }
    
    public final byte getByte(Object o, long offset) {
        return this.unsafe.getByte(o, offset);
    }
    
    public final void putByte(Object o, long offset, byte x) {
        this.unsafe.putByte(o, offset, x);
    }
    
    public final short getShort(Object o, long offset) {
        return this.unsafe.getShort(o, offset);
    }
    
    public final void putShort(Object o, long offset, short x) {
        this.unsafe.putShort(o, offset, x);
    }
    
    public final char getChar(Object o, long offset) {
        return this.unsafe.getChar(o, offset);
    }
    
    public final void putChar(Object o, long offset, char x) {
        this.unsafe.putChar(o, offset, x);
    }
    
    public final long getLong(Object o, long offset) {
        return this.unsafe.getLong(o, offset);
    }
    
    public final void putLong(Object o, long offset, long x) {
        this.unsafe.putLong(o, offset, x);
    }
    
    public final float getFloat(Object o, long offset) {
        return this.unsafe.getFloat(o, offset);
    }
    
    public final void putFloat(Object o, long offset, float x) {
        this.unsafe.putFloat(o, offset, x);
    }
    
    public final double getDouble(Object o, long offset) {
        return this.unsafe.getDouble(o, offset);
    }
    
    public final void putDouble(Object o, long offset, double x) {
        this.unsafe.putDouble(o, offset, x);
    }
    
    public final long objectFieldOffset(Field f) {
        return this.unsafe.objectFieldOffset(f);
    }
    
    public final void throwException(Throwable ee) {
        this.unsafe.throwException(ee);
    }
    
    public final Constructor newConstructorForSerialization(Class cl, Constructor constructorToCall) {
        return this.reflectionFactory.newConstructorForSerialization(cl, constructorToCall);
    }
    
    static {
        NO_ARGS = new Class[0];
        getBridgePermission = new BridgePermission("getBridge");
        Bridge.bridge = null;
    }
}
